#include <tcl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "uttk.h"

/* environmentCheck
 *
 * Checks the desktop environment.
 *
 * Results:
 *     an integer
 *     1 - KDE
 *     2 - GNOME
 *     0 - can not say
 *
 * Side effects:
 *     None.
 */

static int
environmentCheck (void)
{
    /* Get environment variables characteristic to KDE/GNOME */
    char *KDE = getenv("KDE_FULL_SESSION");           
    char *GNOME = getenv("GNOME_DESKTOP_SESSION_ID"); 

    /* For KDE we check if we have full session (variable == TRUE). */
    if((KDE != NULL) && (strcmp(KDE, "TRUE") == 0)) {
        return 1;
    }

    /* For GNOME this variable is deprecated... :( */
    if(GNOME != NULL) {
        return 2;
    }

    /* Everything failed! :( */
    return 0;
}

/* Environment
 *
 * Implements the ::uttk::environment command.
 *
 * Results:
 *     Standard Tcl result.
 *
 * Side effects:
 *     Result is set to the detected desktop environment name.
 */

static int
Environment (clientData, interp, argc, argv)
	ClientData clientData;  /* Not used. */
    Tcl_Interp *interp;     /* Current interpreter */
    int argc;               /* Number of arguments */
    char *argv[];           /* Argument strings */
{
    if(environmentCheck() == 1) {
        Tcl_SetResult(interp, "KDE", TCL_STATIC); 
        return TCL_OK;
    }

    if(environmentCheck() == 2) {
        Tcl_SetResult(interp, "GNOME", TCL_STATIC);
        return TCL_OK;
    }

    /* Everything failed! :( */
    Tcl_SetResult(interp, "?!", TCL_STATIC);
    return TCL_ERROR;
}

/* Uttk_Init
 *
 *     Initializes the package and loads the apropriate package - tileqt if 
 *     desktop environment is KDE or tilegtk if it is GNOME. 
 *
 * Results:
 *     Standard Tcl result.
 *
 * Side effects:
 *     The uttk package, namespace and endable are created.
 *     New commands added to the interpreter:
 *         ::uttk::environment
 *     The apropriate theme is loaded.
 */

int
Uttk_Init (interp)
    Tcl_Interp *interp;     /* Current interpreter */
{
	/* We use Tcl_InitStubs to make the extension work with any version of 
	 * the interpreter.
	 */
	if (Tcl_InitStubs(interp, TCL_VERSION, 0) == NULL) {
	    return TCL_ERROR;
	}

	if (Tcl_PkgProvide(interp, "uttk", PACKAGE_VERSION) != TCL_OK) {
		return TCL_ERROR;
	}

	Tcl_Eval(interp, "namespace eval " NS " {}");

	Tcl_CreateObjCommand(interp, NS "::environment", Environment, NULL, NULL);

	int major;
	int minor;

	Tcl_GetVersion(&major, &minor, NULL, NULL);
	if (8 <= major && 5 <= minor) {
	    /* Tcl_Export(interp, namespace, "*", 0);
	     * Tcl_CreateEnsemble(interp, NS, namespace, 0); */
		Tcl_Eval(interp, 
		    "namespace eval " NS " {\n"
		    "namespace export *\n"
		    "namespace ensemble create\n"
		    "}");
	}
    
    /*
     * Get results of Environment function, to know which package is require to 
     * load.
     */
    
    int e = environmentCheck();
    if (e == 1) {
        /* ensures that a suitable version of the tileqt package is loaded into
         * the interpreter */
        Tcl_PkgRequire(interp, "ttk::theme::tileqt", "0.2", 0);
    }

    if (e == 2) { 
        /* ensures that a suitable version of the tileqt package is loaded into 
         * the interpreter */
        Tcl_PkgRequire(interp, "ttk::theme::tilegtk", "0.2", 0);
    }

    if (e == 0) { 
        fprintf(stderr, "None desktop environment found. Using classic Ttk.\n");

    }


    return TCL_OK;
}
